import nibabel as nib
import os.path as op
import os
import numpy as np
import logging

import pimms
from AFQ.tasks.decorators import as_file
from AFQ.tasks.utils import with_name, str_to_desc
import AFQ.data.fetch as afd
from AFQ.utils.path import drop_extension, write_json
from AFQ.definitions.mapping import SynMap
from AFQ.definitions.utils import Definition
from AFQ.definitions.image import ImageDefinition

from dipy.io.streamline import load_tractogram
from dipy.io.stateful_tractogram import Space


logger = logging.getLogger('AFQ')


@pimms.calc("b0_warped")
@as_file('_space-template_desc-b0_dwi.nii.gz')
def export_registered_b0(data_imap, mapping):
    """
    full path to a nifti file containing
    b0 transformed to template space
    """
    mean_b0 = nib.load(data_imap["b0"]).get_fdata()
    warped_b0 = mapping.transform(mean_b0)
    warped_b0 = nib.Nifti1Image(warped_b0, data_imap["reg_template"].affine)
    return warped_b0, dict(b0InSubject=data_imap["b0"])


@pimms.calc("template_xform")
@as_file('_space-subject_desc-template_dwi.nii.gz')
def template_xform(mapping, data_imap):
    """
    full path to a nifti file containing
    registration template transformed to subject space
    """
    template_xform = mapping.transform_inverse(
        data_imap["reg_template"].get_fdata())
    template_xform = nib.Nifti1Image(template_xform, data_imap["dwi_affine"])
    return template_xform, dict()


@pimms.calc("rois")
def export_rois(base_fname, output_dir, data_imap, mapping):
    """
    dictionary of full paths to Nifti1Image files of ROIs
    transformed to subject space
    """
    bundle_dict = data_imap["bundle_dict"]
    rois_dir = op.join(output_dir, 'ROIs')
    os.makedirs(rois_dir, exist_ok=True)
    roi_files = {}
    base_roi_fname = op.join(rois_dir, op.split(base_fname)[1])
    for bundle_name in bundle_dict:
        roi_files[bundle_name] = []
        for roi_fname in bundle_dict.transform_rois(
                bundle_name, mapping, data_imap["dwi_affine"],
                base_fname=base_roi_fname):
            logger.info(f"Saving {roi_fname}")
            roi_files[bundle_name].append(roi_fname)
            meta = {}
            meta_fname = f'{drop_extension(roi_fname)}.json'
            write_json(meta_fname, meta)
    return {'rois': roi_files}


@pimms.calc("mapping")
def mapping(base_fname, dwi_data_file, reg_subject, data_imap,
            mapping_definition=None):
    """
    mapping from subject to template space.

    Parameters
    ----------
    mapping_definition : instance of `AFQ.definitions.mapping`, optional
        This defines how to either create a mapping from
        each subject space to template space or load a mapping from
        another software. If creating a map, will register reg_subject and
        reg_template.
        If None, use SynMap()
        Default: None
    """
    reg_template = data_imap["reg_template"]
    if mapping_definition is None:
        mapping_definition = SynMap()
    if not isinstance(mapping_definition, Definition):
        raise TypeError(
            "mapping must be a mapping defined"
            + " in `AFQ.definitions.mapping`")
    return mapping_definition.get_for_subses(
        base_fname, data_imap["dwi"], dwi_data_file,
        reg_subject, reg_template)


@pimms.calc("mapping")
def sls_mapping(base_fname, dwi_data_file, reg_subject, data_imap,
                tractography_imap, mapping_definition=None):
    """
    mapping from subject to template space.

    Parameters
    ----------
    mapping_definition : instance of `AFQ.definitions.mapping`, optional
        This defines how to either create a mapping from
        each subject space to template space or load a mapping from
        another software. If creating a map, will register reg_subject and
        reg_template.
        If None, use SynMap()
        Default: None
    """
    reg_template = data_imap["reg_template"]
    if mapping_definition is None:
        mapping_definition = SynMap()
    if not isinstance(mapping_definition, Definition):
        raise TypeError(
            "mapping must be a mapping defined"
            + " in `AFQ.definitions.mapping`")
    streamlines_file = tractography_imap["streamlines"]
    tg = load_tractogram(
        streamlines_file, reg_subject,
        Space.VOX, bbox_valid_check=False)
    tg.to_rasmm()

    atlas_fname = op.join(
        afd.afq_home,
        'hcp_atlas_16_bundles',
        'Atlas_in_MNI_Space_16_bundles',
        'whole_brain',
        'whole_brain_MNI.trk')
    if not op.exists(atlas_fname):
        afd.fetch_hcp_atlas_16_bundles()
    hcp_atlas = load_tractogram(
        atlas_fname,
        'same', bbox_valid_check=False)
    return mapping_definition.get_for_subses(
        base_fname, data_imap["dwi"],
        dwi_data_file,
        reg_subject, reg_template,
        subject_sls=tg.streamlines,
        template_sls=hcp_atlas.streamlines)


@pimms.calc("reg_subject")
def get_reg_subject(data_imap,
                    reg_subject_spec="power_map"):
    """
    Nifti1Image which represents this subject
    when registering the subject to the template

    Parameters
    ----------
    reg_subject_spec : str, instance of `AFQ.definitions.ImageDefinition`, optional  # noqa
        The source image data to be registered.
        Can either be a Nifti1Image, an ImageFile, or str.
        if "b0", "dti_fa_subject", "subject_sls", or "power_map,"
        image data will be loaded automatically.
        If "subject_sls" is used, slr registration will be used
        and reg_template should be "hcp_atlas".
        Default: "power_map"
    """
    if not isinstance(reg_subject_spec, str)\
            and not isinstance(reg_subject_spec, nib.Nifti1Image):
        # Note the ImageDefinition case is handled in get_mapping_plan
        raise TypeError(
            "reg_subject must be a str, ImageDefinition, or Nifti1Image")

    filename_dict = {
        "b0": "b0",
        "power_map": "csd_pmap",
        "dti_fa_subject": "dti_fa",
        "subject_sls": "b0",
    }
    bm = nib.load(data_imap["brain_mask"])

    if reg_subject_spec in filename_dict:
        reg_subject_spec = data_imap[filename_dict[reg_subject_spec]]
    if isinstance(reg_subject_spec, str):
        img = nib.load(reg_subject_spec)
    bm = bm.get_fdata().astype(bool)
    masked_data = img.get_fdata()
    masked_data[~bm] = 0
    img = nib.Nifti1Image(masked_data, img.affine)
    return img


def get_mapping_plan(kwargs, use_sls=False):
    mapping_tasks = with_name([
        export_registered_b0, template_xform, export_rois, mapping,
        get_reg_subject])

    # add custom scalars
    for scalar in kwargs["scalars"]:
        if isinstance(scalar, Definition):
            mapping_tasks[f"{scalar.get_name()}_res"] =\
                pimms.calc(f"{scalar.get_name()}")(
                    as_file((
                        f'_desc-{str_to_desc(scalar.get_name())}'
                        '_dwi.nii.gz'))(
                            scalar.get_image_getter("mapping")))

    if use_sls:
        mapping_tasks["mapping_res"] = sls_mapping

    reg_ss = kwargs.get("reg_subject_spec", None)
    if isinstance(reg_ss, ImageDefinition):
        del kwargs["reg_subject_spec"]
        mapping_tasks["reg_subject_spec_res"] = pimms.calc("reg_subject_spec")(
            as_file((
                f'_desc-{str_to_desc(reg_ss.get_name())}'
                '_dwi.nii.gz'))(reg_ss.get_image_getter("mapping")))

    return pimms.plan(**mapping_tasks)



import logging
from collections.abc import MutableMapping, Mapping

import AFQ.data.fetch as afd
import AFQ.utils.volume as auv
from AFQ.tasks.utils import get_fname, str_to_desc
from AFQ.definitions.utils import find_file

import numpy as np
import nibabel as nib

from dipy.io.streamline import load_tractogram

logging.basicConfig(level=logging.INFO)


__all__ = [
    "BundleDict",
    "default18_bd", "reco_bd",
    "callosal_bd", "cerebellar_bd",
    "baby_bd"]


def do_preprocessing():
    raise NotImplementedError


def append_l_r(bundle_list, no_lr_list):
    new_bundle_list = []
    for bundle in bundle_list:
        if bundle in no_lr_list:
            new_bundle_list.append(bundle)
        else:
            new_bundle_list.append(bundle + "_L")
            new_bundle_list.append(bundle + "_R")
    return new_bundle_list


RECO_UNIQUE = [
    'CCMid', 'CC_ForcepsMajor', 'CC_ForcepsMinor', 'MCP', 'AC', 'PC', 'SCP',
    'V', 'CC', 'F_L_R']

RECO_BUNDLES_16 = [
    'CST', 'C', 'F', 'UF', 'MCP', 'AF', 'CCMid',
    'CC_ForcepsMajor', 'CC_ForcepsMinor', 'IFOF']
RECO_BUNDLES_16 = append_l_r(RECO_BUNDLES_16, RECO_UNIQUE)

RECO_BUNDLES_80 = ["AC", "AF", "AR", "AST", "C", "CB", "CC_ForcepsMajor",
                   "CC_ForcepsMinor", "CC", "CCMid", "CNII", "CNIII",
                   "CNIV", "CNV", "CNVII", "CNVIII", "CS", "CST", "CT",
                   "CTT", "DLF", "EMC", "F_L_R", "FPT", "ICP", "IFOF", "ILF",
                   "LL", "MCP", "MdLF", "ML", "MLF", "OPT", "OR", "PC", "PPT",
                   "RST", "SCP", "SLF", "STT", "TPT", "UF", "V", "VOF"]
RECO_BUNDLES_80 = append_l_r(RECO_BUNDLES_80, RECO_UNIQUE)
# See: https://www.cmu.edu/dietrich/psychology/cognitiveaxon/documents/yeh_etal_2018.pdf  # noqa


DIPY_GH = "https://github.com/dipy/dipy/blob/master/dipy/"


def default18_bd():
    templates = afd.read_templates(as_img=False)
    templates['ARC_roi1_L'] = templates['SLF_roi1_L']
    templates['ARC_roi1_R'] = templates['SLF_roi1_R']
    templates['ARC_roi2_L'] = templates['SLFt_roi2_L']
    templates['ARC_roi2_R'] = templates['SLFt_roi2_R']
    callosal_templates =\
        afd.read_callosum_templates(as_img=False)
    return BundleDict({
        'Left Anterior Thalamic': {
            'cross_midline': False,
            'include': [
                templates['ATR_roi1_L'],
                templates['ATR_roi2_L']],
            'exclude': [],
            'space': 'template',
            'prob_map': templates['ATR_L_prob_map'],
            'start': templates['ATR_L_start'],
            'end': templates['ATR_L_end']},
        'Right Anterior Thalamic': {
            'cross_midline': False,
            'include': [templates['ATR_roi1_R'],
                        templates['ATR_roi2_R']],
            'exclude': [],
            'space': 'template',
            'prob_map': templates['ATR_R_prob_map'],
            'start': templates['ATR_R_start'],
            'end': templates['ATR_R_end']},
        'Left Cingulum Cingulate': {
            'cross_midline': False,
            'include': [templates['CGC_roi2_L'],
                        templates['CGC_roi1_L']],
            'exclude': [],
            'space': 'template',
            'prob_map': templates['CGC_L_prob_map'],
            'end': templates['CGC_L_start']},
        'Right Cingulum Cingulate': {
            'cross_midline': False,
            'include': [templates['CGC_roi2_R'],
                        templates['CGC_roi1_R']],
            'exclude': [],
            'space': 'template',
            'prob_map': templates['CGC_R_prob_map'],
            'end': templates['CGC_R_start']},
        'Left Corticospinal': {
            'cross_midline': False,
            'include': [templates['CST_roi2_L'],
                        templates['CST_roi1_L']],
            'exclude': [],
            'space': 'template',
            'prob_map': templates['CST_L_prob_map'],
            'end': templates['CST_L_start'],
            'start': templates['CST_L_end']},
        'Right Corticospinal': {
            'cross_midline': False,
            'include': [templates['CST_roi2_R'],
                        templates['CST_roi1_R']],
            'exclude': [],
            'space': 'template',
            'prob_map': templates['CST_R_prob_map'],
            'end': templates['CST_R_start'],
            'start': templates['CST_R_end']},
        'Left Inferior Fronto-occipital': {
            'cross_midline': False,
            'include': [templates['IFO_roi2_L'],
                        templates['IFO_roi1_L']],
            'exclude': [templates['ARC_roi1_L']],
            'space': 'template',
            'prob_map': templates['IFO_L_prob_map'],
            'end': templates['IFO_L_start'],
            'start': templates['IFO_L_end']},
        'Right Inferior Fronto-occipital': {
            'cross_midline': False,
            'include': [templates['IFO_roi2_R'],
                        templates['IFO_roi1_R']],
            'exclude': [templates['ARC_roi1_R']],
            'space': 'template',
            'prob_map': templates['IFO_R_prob_map'],
            'end': templates['IFO_R_start'],
            'start': templates['IFO_R_end']},
        'Left Inferior Longitudinal': {
            'cross_midline': False,
            'include': [templates['ILF_roi2_L'],
                        templates['ILF_roi1_L']],
            'exclude': [],
            'space': 'template',
            'prob_map': templates['ILF_L_prob_map'],
            'start': templates['ILF_L_end'],
            'end': templates['ILF_L_start']},
        'Right Inferior Longitudinal': {
            'cross_midline': False,
            'include': [templates['ILF_roi2_R'],
                        templates['ILF_roi1_R']],
            'exclude': [],
            'space': 'template',
            'prob_map': templates['ILF_R_prob_map'],
            'start': templates['ILF_R_end'],
            'end': templates['ILF_R_start']},
        'Left Superior Longitudinal': {
            'cross_midline': False,
            'include': [templates['SLF_roi1_L'],
                        templates['SLF_roi2_L']],
            'exclude': [templates['SLFt_roi2_L']],
            'space': 'template',
            'prob_map': templates['SLF_L_prob_map'],
            'start': templates['SLF_L_start'],
            'end': templates['SLF_L_end']},
        'Right Superior Longitudinal': {
            'cross_midline': False,
            'include': [templates['SLF_roi1_R'],
                        templates['SLF_roi2_R']],
            'exclude': [templates['SLFt_roi2_R']],
            'space': 'template',
            'prob_map': templates['SLF_R_prob_map'],
            'start': templates['SLF_R_start'],
            'end': templates['SLF_R_end']},
        'Left Arcuate': {'cross_midline': False,
                         'include': [templates['SLF_roi1_L'],
                                     templates['SLFt_roi2_L']],
                         'exclude': [],
                         'space': 'template',
                         'prob_map': templates['ARC_L_prob_map'],
                         'start': templates['ARC_L_start'],
                         'end': templates['ARC_L_end']},
        'Right Arcuate': {'cross_midline': False,
                          'include': [templates['SLF_roi1_R'],
                                      templates['SLFt_roi2_R']],
                          'exclude': [],
                          'space': 'template',
                          'prob_map': templates['ARC_R_prob_map'],
                          'start': templates['ARC_R_start'],
                          'end': templates['ARC_R_end']},
        'Left Uncinate': {'cross_midline': False,
                          'include': [templates['UNC_roi2_L'],
                                      templates['UNC_roi1_L']],
                          'exclude': [],
                          'space': 'template',
                          'prob_map': templates['UNC_L_prob_map'],
                          'start': templates['UNC_L_end'],
                          'end': templates['UNC_L_start']},
        'Right Uncinate': {'cross_midline': False,
                           'include': [templates['UNC_roi2_R'],
                                       templates['UNC_roi1_R']],
                           'exclude': [],
                           'space': 'template',
                           'prob_map': templates['UNC_R_prob_map'],
                           'start': templates['UNC_R_end'],
                           'end': templates['UNC_R_start']},
        'Forceps Minor': {'cross_midline': True,
                          'include': [templates['FA_L'],
                                      callosal_templates['Callosum_midsag'],
                                      templates['FA_R']],
                          'exclude': [],
                          'space': 'template',
                          'prob_map': templates['FA_prob_map'],
                          'start': templates['FA_start'],
                          'end': templates['FA_end']},
        'Forceps Major': {'cross_midline': True,
                          'include': [templates['FP_L'],
                                      callosal_templates['Callosum_midsag'],
                                      templates['FP_R']],
                          'exclude': [],
                          'space': 'template',
                          'prob_map': templates['FP_prob_map'],
                          'start': templates['FP_start'],
                          'end': templates['FP_end']},
        'Left Posterior Arcuate': {'cross_midline': False,
                                   'include': [templates['SLFt_roi2_L']],
                                   'exclude': [templates['SLF_roi1_L']],
                                   'space': 'template',
                                   'start': templates['pARC_L_start'],
                                   'primary_axis': 2,
                                   'primary_axis_percentage': 40},
        'Right Posterior Arcuate': {'cross_midline': False,
                                    'include': [templates['SLFt_roi2_R']],
                                    'exclude': [templates['SLF_roi1_R']],
                                    'space': 'template',
                                    'start': templates['pARC_R_start'],
                                    'primary_axis': 2,
                                    'primary_axis_percentage': 40},
        'Left Vertical Occipital': {'cross_midline': False,
                                    'space': 'template',
                                    'start': templates['VOF_L_start'],
                                    'end': templates['VOF_L_end'],
                                    'inc_addtol': [4, 0],
                                    'primary_axis': 2,
                                    'primary_axis_percentage': 40},
        'Right Vertical Occipital': {'cross_midline': False,
                                     'space': 'template',
                                     'start': templates['VOF_R_start'],
                                     'end': templates['VOF_R_end'],
                                     'inc_addtol': [4, 0],
                                     'primary_axis': 2,
                                     'primary_axis_percentage': 40}})


def baby_bd():
    # Pediatric bundles differ from adult bundles:
    #   - A third ROI has been introduced for curvy tracts:
    #     ARC, ATR, CGC, IFO, and UCI
    #   - ILF posterior ROI has been split into two
    #     to separate ILF and mdLF
    #   - Addition of pAF and VOF ROIs
    #   - SLF ROIs are restricted to parietal cortex
    templates = afd.read_pediatric_templates()

    # pediatric probability maps
    prob_map_order = [
        "ATR_L", "ATR_R", "CST_L", "CST_R", "CGC_L", "CGC_R",
        "MdLF_L", "MdLF_R", "FP", "FA", "IFO_L", "IFO_R", "ILF_L",
        "ILF_R", "SLF_L", "SLF_R", "UNC_L", "UNC_R",
        "ARC_L", "ARC_R"]

    prob_maps = templates[
        'UNCNeo_JHU_tracts_prob-for-babyAFQ']
    prob_map_data = prob_maps.get_fdata()

    for bundle_name in prob_map_order:
        templates[bundle_name + "_prob_map"] = nib.Nifti1Image(
            prob_map_data[
                ...,
                prob_map_order.index(bundle_name)], prob_maps.affine)

    # reuse probability map from ILF
    templates["MdLF_L_prob_map"] = templates["ILF_L_prob_map"]
    templates["MdLF_R_prob_map"] = templates["ILF_R_prob_map"]
    return BundleDict({
        'Left Anterior Thalamic': {
            'cross_midline': False,
            'include': [
                templates['ATR_roi3_L'],
                templates['ATR_roi2_L'],
                templates['ATR_roi1_L']],
            'exclude': [],
            'space': 'template',
            'prob_map': templates['ATR_L_prob_map'],
            'mahal': {'distance_threshold': 4}},
        'Right Anterior Thalamic': {'cross_midline': False,
                                    'include': [
                                        templates['ATR_roi3_R'],
                                        templates['ATR_roi2_R'],
                                        templates['ATR_roi1_R']],
                                    'exclude': [],
                                    'space': 'template',
                                    'prob_map': templates['ATR_R_prob_map'],
                                    'mahal': {'distance_threshold': 4}},
        'Left Cingulum Cingulate': {
            'cross_midline': False,
            'include': [templates['CGC_roi3_L'],
                        templates['CGC_roi2_L'],
                        templates['CGC_roi1_L']],
            'exclude': [],
            'space': 'template',
            'prob_map': templates['CGC_L_prob_map'],
            'mahal': {'distance_threshold': 4}},
        'Right Cingulum Cingulate': {
            'cross_midline': False,
            'include': [templates['CGC_roi3_R'],
                        templates['CGC_roi2_R'],
                        templates['CGC_roi1_R']],
            'exclude': [],
            'space': 'template',
            'prob_map': templates['CGC_R_prob_map'],
            'mahal': {'distance_threshold': 4}},
        'Left Corticospinal': {'cross_midline': False,
                               'include': [templates['CST_roi2_L'],
                                           templates['CST_roi1_L']],
                               'exclude': [],
                               'space': 'template',
                               'prob_map': templates['CST_L_prob_map'],
                               'mahal': {'distance_threshold': 4}},
        'Right Corticospinal': {'cross_midline': False,
                                'include': [templates['CST_roi2_R'],
                                            templates['CST_roi1_R']],
                                'exclude': [],
                                'space': 'template',
                                'prob_map': templates['CST_R_prob_map'],
                                'mahal': {'distance_threshold': 4}},
        'Left Inferior Fronto-occipital': {
            'cross_midline': False,
            'include': [templates['IFO_roi3_L'],
                        templates['IFO_roi2_L'],
                        templates['IFO_roi1_L']],
            'exclude': [],
            'space': 'template',
            'prob_map': templates['IFO_L_prob_map'],
            'mahal': {'distance_threshold': 4}},
        'Right Inferior Fronto-occipital': {
            'cross_midline': False,
            'include': [templates['IFO_roi3_R'],
                        templates['IFO_roi2_R'],
                        templates['IFO_roi1_R']],
            'exclude': [],
            'space': 'template',
            'prob_map': templates['IFO_R_prob_map'],
            'mahal': {'distance_threshold': 4}},
        'Left Inferior Longitudinal': {
            'cross_midline': False,
            'include': [templates['ILF_roi2_L'],
                        templates['ILF_roi1_L']],
            'exclude': [],
            'space': 'template',
            'prob_map': templates['ILF_L_prob_map'],
            'mahal': {'distance_threshold': 4}},
        'Right Inferior Longitudinal': {
            'cross_midline': False,
            'include': [templates['ILF_roi2_R'],
                        templates['ILF_roi1_R']],
            'exclude': [],
            'space': 'template',
            'prob_map': templates['ILF_R_prob_map'],
            'mahal': {'distance_threshold': 4}},
        'Left Middle Longitudinal': {
            'cross_midline': False,
            'include': [templates['MdLF_roi2_L'],
                        templates['MdLF_roi1_L']],
            'exclude': [],
            'space': 'template',
            'prob_map': templates['MdLF_L_prob_map'],
            'mahal': {'distance_threshold': 4}},
        'Right Middle Longitudinal': {
            'cross_midline': False,
            'include': [templates['MdLF_roi2_R'],
                        templates['MdLF_roi1_R']],
            'exclude': [],
            'space': 'template',
            'prob_map': templates['MdLF_R_prob_map'],
            'mahal': {'distance_threshold': 4}},
        'Left Superior Longitudinal': {
            'cross_midline': False,
            'include': [templates['SLF_roi1_L'],
                        templates['SLF_roi2_L']],
            'exclude': [templates['SLFt_roi2_L']],
            'space': 'template',
            'prob_map': templates['SLF_L_prob_map'],
            'mahal': {'distance_threshold': 4}},
        'Right Superior Longitudinal': {
            'cross_midline': False,
            'include': [templates['SLF_roi1_R'],
                        templates['SLF_roi2_R']],
            'exclude': [templates['SLFt_roi2_R']],
            'space': 'template',
            'prob_map': templates['SLF_R_prob_map'],
            'mahal': {'distance_threshold': 4}},
        'Left Arcuate': {'cross_midline': False,
                         'include': [templates['ARC_roi1_L'],
                                     templates['ARC_roi2_L'],
                                     templates['ARC_roi3_L']],
                         'exclude': [],
                         'space': 'template',
                         'prob_map': templates['ARC_L_prob_map'],
                         'mahal': {'distance_threshold': 4}},
        'Right Arcuate': {'cross_midline': False,
                          'include': [templates['ARC_roi1_R'],
                                      templates['ARC_roi2_R'],
                                      templates['ARC_roi3_R']],
                          'exclude': [],
                          'space': 'template',
                          'prob_map': templates['ARC_R_prob_map'],
                          'mahal': {'distance_threshold': 4}},
        'Left Uncinate': {'cross_midline': False,
                          'include': [templates['UNC_roi3_L'],
                                      templates['UNC_roi2_L'],
                                      templates['UNC_roi1_L']],
                          'exclude': [],
                          'space': 'template',
                          'prob_map': templates['UNC_L_prob_map'],
                          'mahal': {'distance_threshold': 4}},
        'Right Uncinate': {'cross_midline': False,
                           'include': [templates['UNC_roi3_R'],
                                       templates['UNC_roi2_R'],
                                       templates['UNC_roi1_R']],
                           'exclude': [],
                           'space': 'template',
                           'prob_map': templates['UNC_R_prob_map'],
                           'mahal': {'distance_threshold': 4}},
        'Forceps Minor': {'cross_midline': True,
                          'include': [templates['FA_R'],
                                      templates['mid-saggital'],
                                      templates['FA_L']],
                          'exclude': [],
                          'space': 'template',
                          'prob_map': templates['FA_prob_map'],
                          'mahal': {'distance_threshold': 4}},
        'Forceps Major': {'cross_midline': True,
                          'include': [templates['FP_R'],
                                      templates['mid-saggital'],
                                      templates['FP_L']],
                          'exclude': [],
                          'space': 'template',
                          'prob_map': templates['FP_prob_map'],
                          'mahal': {'distance_threshold': 4}},
        'Left Optic Radiation': {
            "include": [templates["OR_left_roi3"]],
            "start": templates["OR_leftThal"],
            "end": templates["OR_leftV1"],
            "cross_midline": False,
            "mahal": {"distance_threshold": 4}},
        'Right Optic Radiation': {
            "include": [templates["OR_right_roi3"]],
            "start": templates["OR_rightThal"],
            "end": templates["OR_rightV1"],
            "cross_midline": False,
            "mahal": {"distance_threshold": 4}},
        'Left Posterior Arcuate': {
            "include": [templates["SLFt_roi2_L"]],
            "exclude": [templates["SLF_roi1_L"]],
            "start": templates["pARC_L_start"],
            "end": templates["VOF_box_small_L"],
            "primary_axis": 2,
            "primary_axis_percentage": 40,
            "cross_midline": False,
            "mahal": {"distance_threshold": 4}},
        'Right Posterior Arcuate': {
            "include": [templates["SLFt_roi2_R"]],
            "exclude": [templates["SLF_roi1_R"]],
            "start": templates["pARC_R_start"],
            "end": templates["VOF_box_small_R"],
            "primary_axis": 2,
            "primary_axis_percentage": 40,
            "cross_midline": False,
            "mahal": {"distance_threshold": 4}},
        'Left Vertical Occipital': {
            "start": templates["VOF_L_start"],
            "end": templates["VOF_box_small_L"],
            "primary_axis": 2,
            "primary_axis_percentage": 40,
            "cross_midline": False,
            "mahal": {"distance_threshold": 4}},
        'Right Vertical Occipital': {
            "start": templates["VOF_R_start"],
            "end": templates["VOF_box_small_R"],
            "primary_axis": 2,
            "primary_axis_percentage": 40,
            "cross_midline": False,
            "mahal": {"distance_threshold": 4}}},
        resample_to=afd.read_pediatric_templates()[
            'UNCNeo-withCerebellum-for-babyAFQ'])


def callosal_bd():
    callosal_templates =\
        afd.read_callosum_templates(as_img=False)
    return BundleDict({
        'Callosum Anterior Frontal': {
            'cross_midline': True,
            'include': [callosal_templates['R_AntFrontal'],
                        callosal_templates['Callosum_midsag'],
                        callosal_templates['L_AntFrontal']],
            'exclude': [],
            'space': 'template'},
        'Callosum Motor': {
            'cross_midline': True,
            'include': [callosal_templates['R_Motor'],
                        callosal_templates['Callosum_midsag'],
                        callosal_templates['L_Motor']],
            'exclude': [],
            'space': 'template'},
        'Callosum Occipital': {
            'cross_midline': True,
            'include': [callosal_templates['R_Occipital'],
                        callosal_templates['Callosum_midsag'],
                        callosal_templates['L_Occipital']],
            'exclude': [],
            'space': 'template'},
        'Callosum Orbital': {
            'cross_midline': True,
            'include': [callosal_templates['R_Orbital'],
                        callosal_templates['Callosum_midsag'],
                        callosal_templates['L_Orbital']],
            'exclude': [],
            'space': 'template'},
        'Callosum Posterior Parietal': {
            'cross_midline': True,
            'include': [callosal_templates['R_PostParietal'],
                        callosal_templates['Callosum_midsag'],
                        callosal_templates['L_PostParietal']],
            'exclude': [],
            'space': 'template'},
        'Callosum Superior Frontal': {
            'cross_midline': True,
            'include': [callosal_templates['R_SupFrontal'],
                        callosal_templates['Callosum_midsag'],
                        callosal_templates['L_SupFrontal']],
            'exclude': [],
            'space': 'template'},
        'Callosum Superior Parietal': {
            'cross_midline': True,
            'include': [callosal_templates['R_SupParietal'],
                        callosal_templates['Callosum_midsag'],
                        callosal_templates['L_SupParietal']],
            'exclude': [],
            'space': 'template'},
        'Callosum Temporal': {
            'cross_midline': True,
            'include': [callosal_templates['R_Temporal'],
                        callosal_templates['Callosum_midsag'],
                        callosal_templates['L_Temporal']],
            'exclude': [],
            'space': 'template'}})


def reco_bd(n_bundles):
    """
    n_bundles: int
        Selects between 16 or 80 bundle atlas
    """
    templates = afd.read_hcp_atlas(n_bundles, as_file=True)
    return BundleDict(templates)


def cerebellar_bd():
    cp_rois = afd.read_cp_templates()
    return BundleDict(
        {
            "Left Inferior Cerebellar Peduncle": {
                "include": [
                    cp_rois["ICP_L_inferior_prob"],
                    cp_rois["ICP_L_superior_prob"],
                ],
                "cross_midline": False,
            },
            "Right Inferior Cerebellar Peduncle": {
                "include": [
                    cp_rois["ICP_R_inferior_prob"],
                    cp_rois["ICP_R_superior_prob"],
                ],
                "cross_midline": False,
            },
            "Left Middle Cerebellar Peduncle": {
                "include": [
                    cp_rois["MCP_L_inferior_prob"],
                    cp_rois["MCP_R_superior_prob"],
                ],
                "exclude": [
                    cp_rois["SCP_L_inter_prob"],
                ],
                "cross_midline": True,
            },
            "Right Middle Cerebellar Peduncle": {
                "include": [
                    cp_rois["MCP_R_inferior_prob"],
                    cp_rois["MCP_L_superior_prob"],
                ],
                "exclude": [
                    cp_rois["SCP_R_inter_prob"],
                ],
                "cross_midline": True,
            },
            "Left Superior Cerebellar Peduncle": {
                "include": [
                    cp_rois["SCP_L_inferior_prob"],
                    cp_rois["SCP_L_inter_prob"],
                    cp_rois["SCP_R_superior_prob"],
                ],
                "exclude": [
                    cp_rois["SCP_L_superior_prob"],
                ],
                "cross_midline": True,
            },
            "Right Superior Cerebellar Peduncle": {
                "include": [
                    cp_rois["SCP_R_inferior_prob"],
                    cp_rois["SCP_R_inter_prob"],
                    cp_rois["SCP_L_superior_prob"],
                ],
                "exclude": [
                    cp_rois["SCP_R_superior_prob"],
                ],
                "cross_midline": True}})


def OR_bd():
    or_rois = afd.read_or_templates()

    return BundleDict({
        "Left Optic Radiation": {
            "include": [
                or_rois["left_OR_1"],
                or_rois["left_OR_2"]],
            "exclude": [
                or_rois["left_OP_MNI"],
                or_rois["left_TP_MNI"],
                or_rois["left_pos_thal_MNI"]],
            "start": or_rois['left_thal_MNI'],
            "end": or_rois['left_V1_MNI'],
            "cross_midline": False,
        },
        "Right Optic Radiation": {
            "include": [
                or_rois["right_OR_1"],
                or_rois["right_OR_2"]],
            "exclude": [
                or_rois["right_OP_MNI"],
                or_rois["right_TP_MNI"],
                or_rois["right_pos_thal_MNI"]],
            "start": or_rois['right_thal_MNI'],
            "end": or_rois['right_V1_MNI'],
            "cross_midline": False
        }
    })


class _BundleEntry(Mapping):
    """Describes how to recognize a single bundle, immutable"""

    def __init__(self, data):
        self._data = data

    def __getitem__(self, key):
        return self._data[key]

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return iter(self._data)

    def __setitem__(self, key, value):
        raise RuntimeError((
            "You cannot modify the properties of a bundle's definition. "
            "To modify a bundle's definition, replace that bundle's entry "
            "in the BundleDict."))


class BundleDict(MutableMapping):
    """
    Create a bundle dictionary, needed for the segmentation.

    Parameters
    ----------
    bundle_info : dict,
        A dictionary defining
        custom bundles. See `Defining Custom Bundle Dictionaries`
        in the `usage` section of pyAFQ's documentation for details.

    resample_to : Nifti1Image or bool, optional
        If there are bundles in bundle_info with the 'space' attribute
        set to 'template', or with no 'space' attribute,
        their images (all ROIs and probability maps)
        will be resampled to the affine and shape of this image.
        If None, the MNI template will be used.
        If False, no resampling will be done.
        Default: None

    resample_subject_to : Nifti1Image or bool, optional
        If there are bundles in bundle_info with the 'space' attribute
        set to 'subject', their images (all ROIs and probability maps)
        will be resampled to the affine and shape of this image.
        If False, no resampling will be done.
        Default: None

    keep_in_memory : bool, optional
        Whether, once loaded, all ROIs and probability maps will stay
        loaded in memory within this object. By default, ROIs are loaded
        into memory on demand and no references to ROIs are kept, other
        than their paths. The default 18 bundles use ~6GB when all loaded.
        Default: False

    Examples
    --------
    # import OR ROIs and create a custom bundle dict
    # from them
    import AFQ.data.fetch as afd
    or_rois = afd.read_or_templates()

    bundles = BundleDict({
        "L_OR": {
            "include": [
                or_rois["left_OR_1"],  # these can be paths to Nifti files
                or_rois["left_OR_2"]],  # or they can Nifti images
            "exclude": [
                or_rois["left_OP_MNI"],
                or_rois["left_TP_MNI"],
                or_rois["left_pos_thal_MNI"]],
            "start": or_rois['left_thal_MNI'],
            "end": or_rois['left_V1_MNI'],
            "cross_midline": False,
        },
        "R_OR": {
            "include": [
                or_rois["right_OR_1"],
                or_rois["right_OR_2"]],
            "exclude": [
                or_rois["right_OP_MNI"],
                or_rois["right_TP_MNI"],
                or_rois["right_pos_thal_MNI"]],
            "start": or_rois['right_thal_MNI'],
            "end": or_rois['right_V1_MNI'],
            "cross_midline": False
        }
    })
    """

    def __init__(self,
                 bundle_info,
                 resample_to=None,
                 resample_subject_to=False,
                 keep_in_memory=False):
        if not (isinstance(bundle_info, dict)):
            raise TypeError((
                f"bundle_info must be a dict,"
                f" currently a {type(bundle_info)}"))
        if resample_to is None:
            resample_to = afd.read_mni_template()
        self.resample_to = resample_to
        self.resample_subject_to = resample_subject_to
        self.keep_in_memory = keep_in_memory
        self.max_includes = 3

        self._dict = {}
        self.bundle_names = []
        for key, item in bundle_info.items():
            self.__setitem__(key, item)

        self.logger = logging.getLogger('AFQ')
        if "Forceps Major" in self.bundle_names\
                and "Callosum Occipital" in self.bundle_names:
            self.logger.info((
                "Forceps Major and Callosum Occipital bundles"
                " are co-located, and AFQ"
                " assigns each streamline to only one bundle."
                " Only Callosum Occipital will be used."))
            self.bundle_names.remove("Forceps Major")
        if "Forceps Minor" in self.bundle_names\
                and "Callosum Orbital" in self.bundle_names:
            self.logger.info((
                "Forceps Minor and Callosum Orbital bundles"
                " are co-located, and AFQ"
                " assigns each streamline to only one bundle."
                " Only Callosum Orbital will be used."))
            self.bundle_names.remove("Forceps Minor")
        if "Forceps Minor" in self.bundle_names\
                and "Callosum Anterior Frontal" in self.bundle_names:
            self.logger.info((
                "Forceps Minor and Callosum Anterior Frontal bundles"
                " are co-located, and AFQ"
                " assigns each streamline to only one bundle."
                " Only Callosum Anterior Frontal will be used."))
            self.bundle_names.remove("Forceps Minor")

    def update_max_includes(self, new_max):
        if new_max > self.max_includes:
            self.max_includes = new_max

    def _use_bids_info(self, roi_or_sl, bids_layout, bids_path,
                       subject, session):
        if isinstance(roi_or_sl, dict):
            suffix = roi_or_sl.get("suffix", "dwi")
            roi_or_sl = find_file(
                bids_layout, bids_path,
                roi_or_sl,
                suffix,
                session, subject)
            return roi_or_sl
        else:
            return roi_or_sl

    def _cond_load(self, roi_or_sl, resample_to):
        """
        Load ROI or streamline if not already loaded
        """
        if isinstance(roi_or_sl, str):
            if ".nii" in roi_or_sl:
                return afd.read_resample_roi(
                    roi_or_sl,
                    resample_to=resample_to)
            else:
                return load_tractogram(
                    roi_or_sl,
                    'same',
                    bbox_valid_check=False).streamlines
        elif isinstance(roi_or_sl, nib.Nifti1Image):
            return afd.read_resample_roi(
                roi_or_sl,
                resample_to=resample_to)
        else:
            return roi_or_sl

    def get_b_info(self, b_name):
        return self._dict[b_name]

    def __getitem__(self, key):
        if isinstance(key, tuple) or isinstance(key, list):
            # Generates a copy of this BundleDict with only the bundle names
            # from the tuple/list
            new_bd = {}
            for b_name in key:
                if b_name in self._dict:
                    new_bd[b_name] = self._dict[b_name]
                else:
                    raise ValueError(f"{b_name} is not in this BundleDict")

            return self.__class__(
                new_bd,
                resample_to=self.resample_to,
                resample_subject_to=self.resample_subject_to,
                keep_in_memory=self.keep_in_memory)
        else:
            if not self.keep_in_memory:
                _item = self._dict[key].copy()
                _res = self._cond_load_bundle(key, dry_run=True)
                if _res is not None:
                    _item.update(_res)
                _item = _BundleEntry(_item)
            else:
                if "loaded" not in self._dict[key] or\
                        not self._dict[key]["loaded"]:
                    self._cond_load_bundle(key)
                    self._dict[key]["loaded"] = True
                if "resampled" not in self._dict[key] or not self._dict[
                        key]["resampled"]:
                    self._resample_roi(key)
                _item = _BundleEntry(self._dict[key].copy())
            return _item

    def __setitem__(self, key, item):
        self._dict[key] = item
        if hasattr(item, "get"):
            self.update_max_includes(len(item.get("include", [])))
        if key not in self.bundle_names:
            self.bundle_names.append(key)

    def __len__(self):
        return len(self.bundle_names)

    def __delitem__(self, key):
        if key not in self._dict and key not in self.bundle_names:
            raise KeyError(f"{key} not found")
        if key in self._dict:
            del self._dict[key]
        else:
            raise RuntimeError((
                f"{key} not found in internal dictionary, "
                f"but found in bundle_names"))
        if key in self.bundle_names:
            self.bundle_names.remove(key)
        else:
            raise RuntimeError((
                f"{key} not found in bundle_names, "
                f"but found in internal dictionary"))

    def __iter__(self):
        return iter(self._dict)

    def copy(self):
        """
        Generates a copy of this BundleDict where the internal dictionary
        is a copy of this BundleDict's internal dictionary.
        Useful if you want to add or remove bundles from a copy
        of a BundleDict.

        Returns
        ---------
        bundle_dict : BundleDict
            Euclidean norms of vectors.
        """
        return self.__class__(
            self._dict.copy(),
            resample_to=self.resample_to,
            resample_subject_to=self.resample_subject_to,
            keep_in_memory=self.keep_in_memory)

    def apply_to_rois(self, b_name, func, *args,
                      dry_run=False, apply_to_recobundles=False,
                      **kwargs):
        """
        Applies some transformation to all ROIs (include, exclude, end, start)
        and the prob_map in a given bundle.

        Parameters
        ----------
        b_name : name
            bundle name of bundle whose ROIs will be transformed.
        func : function
            function whose first argument must be a Nifti1Image and which
            returns a Nifti1Image
        dry_run : bool
            Whether to actually apply changes returned by `func` to the ROIs.
            If has_return is False, dry_run is not used.
        apply_to_recobundles : bool, optional
            Whether to apply the transformation to recobundles
            TRKs as well.
            Default: False

        *args :
            Additional arguments for func
        **kwargs
            Optional arguments for func

        Returns
        -------
        A dictionary where keys are
        the roi type and values are the transformed ROIs.
        """
        return_vals = {}
        for roi_type in [
                "include", "exclude",
                "start", "end", "prob_map"]:
            if roi_type in self._dict[b_name]:
                if roi_type in ["start", "end", "prob_map"]:
                    return_vals[roi_type] = func(
                        self._dict[b_name][roi_type], *args, **kwargs)
                else:
                    changed_rois = []
                    for _roi in self._dict[b_name][roi_type]:
                        changed_rois.append(func(
                            _roi, *args, **kwargs))
                    return_vals[roi_type] = changed_rois
        if apply_to_recobundles and "recobundles" in self._dict[b_name]:
            return_vals["recobundles"] = {}
            for sl_type in ["sl", "centroid"]:
                return_vals["recobundles"][sl_type] = func(
                    self._dict[b_name]["recobundles"][sl_type],
                    *args, **kwargs)
        if not dry_run:
            for roi_type, roi in return_vals.items():
                self._dict[b_name][roi_type] = roi
        return return_vals

    def _cond_load_bundle(self, b_name, dry_run=False):
        """
        Given a bundle name, resample all ROIs and prob maps
        into either template or subject space for that bundle,
        depending on its "space" attribute.

        Parameters
        ----------
        b_name : str
            Name of the bundle to be resampled.
        """
        if self.is_bundle_in_template(b_name):
            resample_to = self.resample_to
        else:
            resample_to = self.resample_subject_to
        return self.apply_to_rois(
            b_name,
            self._cond_load,
            resample_to,
            dry_run=dry_run,
            apply_to_recobundles=True)

    def is_bundle_in_template(self, bundle_name):
        return "space" not in self._dict[bundle_name]\
            or self._dict[bundle_name]["space"] == "template"

    def _roi_transform_helper(self, roi_or_sl, mapping,
                              new_affine, bundle_name):
        roi_or_sl = self._cond_load(roi_or_sl, self.resample_to)
        if isinstance(roi_or_sl, nib.Nifti1Image):
            warped_img = auv.transform_inverse_roi(
                roi_or_sl.get_fdata(),
                mapping,
                bundle_name=bundle_name)
            warped_img = nib.Nifti1Image(warped_img, new_affine)
            return warped_img
        else:
            return roi_or_sl

    def transform_rois(self, bundle_name, mapping, new_affine,
                       base_fname=None, apply_to_recobundles=False):
        """
        Get the bundle definition with transformed ROIs
        for a given bundle into a
        given subject space using a given mapping.
        Will only run on bundles which are in template
        space, otherwise will just return the bundle
        definition without transformation.

        Parameters
        ----------
        bundle_name : str
            Name of the bundle to be transformed.
        mapping : DiffeomorphicMap object
            A mapping between DWI space and a template.
        new_affine : array
            Affine of space transformed into.
        base_fname : str, optional
            Base file path to save ROIs too. Additional BIDS
            descriptors will be added to this file path. If None,
            do not save the ROIs.
        apply_to_recobundles : bool, optional
            Whether to apply the transformation to recobundles
            TRKs as well.
            Default: False

        Returns
        -------
        If base_fname is None, a dictionary where keys are
        the roi type and values are the transformed ROIs.
        Otherwise, a list of file names where the transformed
        ROIs are saved.
        """
        if self.is_bundle_in_template(bundle_name):
            transformed_rois = self.apply_to_rois(
                bundle_name,
                self._roi_transform_helper,
                mapping,
                new_affine,
                bundle_name,
                dry_run=True,
                apply_to_recobundles=apply_to_recobundles)
        else:
            transformed_rois = self.apply_to_rois(
                bundle_name,
                self._cond_load,
                self.resample_subject_to,
                dry_run=True,
                apply_to_recobundles=apply_to_recobundles)

        if base_fname is not None:
            fnames = []
            for roi_type, rois in transformed_rois.items():
                if not isinstance(rois, list):
                    rois = [rois]
                for ii, roi in enumerate(rois):
                    fname = get_fname(
                        base_fname,
                        "_space-subject_desc-"
                        f"{str_to_desc(bundle_name)}{roi_type}{ii}"
                        "_mask.nii.gz")
                    nib.save(
                        nib.Nifti1Image(
                            roi.get_fdata().astype(np.float32),
                            roi.affine), fname)
                    fnames.append(fname)
            return fnames
        else:
            return transformed_rois

    def __add__(self, other):
        for resample in ["resample_to", "resample_subject_to"]:
            if not getattr(self, resample)\
                    or not getattr(other, resample)\
                    or getattr(self, resample) is None\
                    or getattr(other, resample) is None:
                if getattr(self, resample) != getattr(other, resample):
                    raise ValueError((
                        f"Adding BundleDicts where {resample} do not match."
                        f"{resample}'s are {getattr(self, resample)} and "
                        f"{getattr(other, resample)}"))
            else:
                if not np.allclose(
                        getattr(self, resample).affine,
                        getattr(other, resample).affine):
                    raise ValueError((
                        f"Adding BundleDicts where {resample} affines"
                        f" do not match. {resample} affines are"
                        f"{getattr(self, resample).affine} and "
                        f"{getattr(other, resample).affine}"))
                if not np.allclose(
                        getattr(self, resample).header['dim'],
                        getattr(other, resample).header['dim']):
                    raise ValueError((
                        f"Adding BundleDicts where {resample} dimensions"
                        f" do not match. {resample} dimensions are"
                        f"{getattr(self, resample).header['dim']} and "
                        f"{getattr(other, resample).header['dim']}"))
        return self.__class__(
            {**self._dict, **other._dict},
            self.resample_to,
            self.resample_subject_to,
            self.keep_in_memory)



import numpy.testing as npt
import numpy as np
import nibabel as nib
import os.path as op

import AFQ.data.fetch as afd
import AFQ.recognition.curvature as abv
import AFQ.recognition.utils as abu
import AFQ.recognition.cleaning as abc


from dipy.io.stateful_tractogram import StatefulTractogram, Space
import dipy.data.fetcher as fetcher


hardi_dir = op.join(fetcher.dipy_home, "stanford_hardi")
hardi_fdata = op.join(hardi_dir, "HARDI150.nii.gz")
hardi_img = nib.load(hardi_fdata)
file_dict = afd.read_stanford_hardi_tractography()
streamlines = file_dict['tractography_subsampled.trk']
tg = StatefulTractogram(streamlines, hardi_img, Space.RASMM)
tg.to_vox()
streamlines = tg.streamlines


def test_segment_sl_curve():
    sl_disp_0 = abv.sl_curve(streamlines[4], 4)
    npt.assert_array_almost_equal(
        sl_disp_0,
        [[-0.236384, -0.763855, 0.60054],
         [0.232594, -0.867859, -0.439],
         [0.175343, 0.001082, -0.984507]])

    sl_disp_1 = abv.sl_curve(streamlines[2], 4)
    mean_angle_diff = abv.sl_curve_dist(sl_disp_0, sl_disp_1)
    npt.assert_almost_equal(mean_angle_diff, 1.701458, decimal=3)


def test_segment_clip_edges():
    sls = tg.streamlines
    idx = np.arange(len(tg.streamlines))
    accepted_sls = sls[[4, 10, 11]]
    accepted_ix = idx[[4, 10, 11]]
    bundle_roi_dists = np.zeros((len(sls), 3))
    bundle_roi_dists[4, :] = [5, 10, 15]
    bundle_roi_dists[10, :] = [3, 6, 9]
    bundle_roi_dists[11, :] = [10, 10, 10]
    cut_sls = abu.cut_sls_by_dist(
        accepted_sls,
        bundle_roi_dists[accepted_ix],
        [0, 2])
    npt.assert_array_equal(
        cut_sls[0],
        accepted_sls[0][5:15])
    npt.assert_array_equal(
        cut_sls[1],
        accepted_sls[1][3:9])
    npt.assert_array_equal(
        cut_sls[2],
        accepted_sls[2][9:11])


def test_segment_orientation():
    cleaned_idx = \
        abc.clean_by_orientation(streamlines, primary_axis=1)
    npt.assert_equal(np.sum(cleaned_idx), 93)
    cleaned_idx_tol = \
        abc.clean_by_orientation(streamlines, primary_axis=1, tol=50)
    npt.assert_(np.sum(cleaned_idx_tol) < np.sum(cleaned_idx))

    cleaned_idx = \
        abc.clean_by_orientation(streamlines, primary_axis=2)
    cleaned_idx_tol = \
        abc.clean_by_orientation(streamlines, primary_axis=2, tol=33)
    npt.assert_array_equal(cleaned_idx_tol, cleaned_idx)



import contextlib
from importlib import import_module
from AFQ.viz.utils import viz_import_msg_error
import pimms
from funcargparse import FuncArgParser
import logging
import warnings

from dipy.io.stateful_tractogram import set_sft_logger_level


with contextlib.suppress(Exception):  # only works on python 3.9
    from outdated import OutdatedPackageWarning
    warnings.filterwarnings("ignore", category=OutdatedPackageWarning)


__all__ = [
    "methods_descriptors", "kwargs_descriptors", "AFQclass_doc"]


set_sft_logger_level(logging.CRITICAL)


task_modules = ["data", "mapping", "segmentation", "tractography", "viz"]

methods_descriptors = {}
methods_sections = {}
kwargs_descriptors = {}
for task_module in task_modules:
    kwargs_descriptors[task_module] = {}
    for calc_obj in import_module(
            f"AFQ.tasks.{task_module}").__dict__.values():
        if isinstance(calc_obj, pimms.calculation.Calc):
            docstr_parser = FuncArgParser()
            docstr_parser.setup_args(calc_obj.function)
            if len(calc_obj.efferents) > 1:
                eff_descs = docstr_parser.description.split(",")
                if len(eff_descs) != len(calc_obj.efferents):
                    raise NotImplementedError((
                        "If calc method has mutliple outputs, "
                        "their descriptions must be divided by commas."
                        f" {calc_obj.name} has {len(eff_descs)} comma-divided"
                        f"sections but {len(calc_obj.efferents)} outputs"))
                for ii in range(len(calc_obj.efferents)):
                    if eff_descs[ii][0] in [' ', '\n']:
                        eff_descs[ii] = eff_descs[ii][1:]
                    if eff_descs[ii][:3] == "and":
                        eff_descs[ii] = eff_descs[ii][3:]
                    if eff_descs[ii][0] in [' ', '\n']:
                        eff_descs[ii] = eff_descs[ii][1:]
                    methods_descriptors[
                        calc_obj.efferents[ii]] =\
                        eff_descs[ii]
                    methods_sections[calc_obj.efferents[ii]] =\
                        task_module
            else:
                methods_descriptors[
                    calc_obj.efferents[0]] =\
                    docstr_parser.description
                methods_sections[calc_obj.efferents[0]] =\
                    task_module
            for arg, info in docstr_parser.unfinished_arguments.items():
                if "help" in info:
                    default = info["default"] if "default" in info else None
                    kwargs_descriptors[task_module][arg] = dict(
                        desc=info["help"],
                        kind=info["metavar"],
                        default=default)
                if arg not in methods_sections:
                    methods_sections[arg] = task_module


AFQclass_doc = (
    "Here are the arguments you can pass to kwargs,"
    " to customize the tractometry pipeline. They are organized"
    " into 5 sections.\n")
for task_module in task_modules:
    AFQclass_doc = AFQclass_doc + "\n"
    AFQclass_doc = AFQclass_doc +\
        "==========================================================\n"
    AFQclass_doc = AFQclass_doc + task_module.upper() + "\n"
    AFQclass_doc = AFQclass_doc +\
        "==========================================================\n"
    for arg, info in kwargs_descriptors[task_module].items():
        AFQclass_doc = AFQclass_doc + arg + ": " + info["kind"]
        AFQclass_doc = AFQclass_doc + "\n\t"
        AFQclass_doc = AFQclass_doc + info["desc"].replace(
            "\n", "\n\t")
        AFQclass_doc = AFQclass_doc + "\n\n"


valid_exports_string = (
    "Here is a list of valid attributes "
    f"to export: {methods_sections.keys()}")


def check_attribute(attr_name):
    if attr_name == "help":
        print(valid_exports_string)
        return None

    if attr_name in methods_sections:
        if methods_sections[attr_name] == task_modules[-1]:
            return None
        else:
            return f"{methods_sections[attr_name]}_imap"

    raise ValueError(
        f"{attr_name} not found for export. {valid_exports_string}")


def export_all_helper(api_afq_object, xforms, indiv, viz):
    if xforms:
        try:
            api_afq_object.export("b0_warped")
        except Exception as e:
            api_afq_object.logger.warning((
                "Failed to export warped b0. This could be because your "
                "mapping type is only compatible with transformation "
                f"from template to subject space. The error is: {e}"))
        api_afq_object.export("template_xform")

    if indiv:
        api_afq_object.export("indiv_bundles")
        api_afq_object.export("rois")
    api_afq_object.export("sl_counts")
    api_afq_object.export("median_bundle_lengths")
    api_afq_object.export("profiles")

    if viz:
        try:
            api_afq_object.export("tract_profile_plots")
        except ImportError as e:
            plot_err_message = viz_import_msg_error("plot")
            if str(e) != plot_err_message:
                raise
            else:
                api_afq_object.logger.warning(plot_err_message)
        api_afq_object.export("all_bundles_figure")
        api_afq_object.export("indiv_bundles_figures")



import numpy as np
from scipy.spatial.distance import cdist
from scipy.ndimage import binary_dilation


def check_sls_with_inclusion(sls, include_rois, include_roi_tols):
    for sl in sls:
        yield check_sl_with_inclusion(
            sl,
            include_rois,
            include_roi_tols)


def check_sl_with_inclusion(sl, include_rois,
                            include_roi_tols):
    """
    Helper function to check that a streamline is close to a list of
    inclusion ROIS.
    """
    dist = []
    for ii, roi in enumerate(include_rois):
        # Use squared Euclidean distance, because it's faster:
        dist.append(cdist(sl, roi, 'sqeuclidean'))
        if np.min(dist[-1]) > include_roi_tols[ii]:
            # Too far from one of them:
            return False, []
    # Apparently you checked all the ROIs and it was close to all of them
    return True, dist


def check_sl_with_exclusion(sl, exclude_rois,
                            exclude_roi_tols):
    """ Helper function to check that a streamline is not too close to a
    list of exclusion ROIs.
    """
    for ii, roi in enumerate(exclude_rois):
        # Use squared Euclidean distance, because it's faster:
        if np.min(cdist(sl, roi, 'sqeuclidean')) < exclude_roi_tols[ii]:
            return False
    # Either there are no exclusion ROIs, or you are not close to any:
    return True


def clean_by_endpoints(streamlines, target, target_idx, tol=0,
                       flip_sls=None, accepted_idxs=None):
    """
    Clean a collection of streamlines based on an endpoint ROI.
    Filters down to only include items that have their start or end points
    close to the targets.
    Parameters
    ----------
    streamlines : sequence of N by 3 arrays
        Where N is number of nodes in the array, the collection of
        streamlines to filter down to.
    target: Nifti1Image
        Nifti1Image containing a boolean representation of the ROI.
    target_idx: int.
        Index within each streamline to check if within the target region.
        Typically 0 for startpoint ROIs or -1 for endpoint ROIs.
        If using flip_sls, this becomes (len(sl) - this_idx - 1) % len(sl)
    tol : int, optional
        A distance tolerance (in units that the coordinates
        of the streamlines are represented in). Default: 0, which means that
        the endpoint is exactly in the coordinate of the target ROI.
    flip_sls : 1d array, optional
        Length is len(streamlines), whether to flip the streamline.
    accepted_idxs : 1d array, optional
        Boolean array, where entries correspond to eachs streamline,
        and streamlines that pass cleaning will be set to 1.
    Yields
    -------
    boolean array of streamlines that survive cleaning.
    """
    if accepted_idxs is None:
        accepted_idxs = np.zeros(len(streamlines), dtype=np.bool8)

    if flip_sls is None:
        flip_sls = np.zeros(len(streamlines))
    flip_sls = flip_sls.astype(int)

    roi = target.get_fdata()
    if tol > 0:
        roi = binary_dilation(
            roi,
            iterations=tol)

    for ii, sl in enumerate(streamlines):
        this_idx = target_idx
        if flip_sls[ii]:
            this_idx = (len(sl) - this_idx - 1) % len(sl)
        xx, yy, zz = sl[this_idx].astype(int)
        accepted_idxs[ii] = roi[xx, yy, zz]

    return accepted_idxs



import AFQ.api.bundle_dict as abd
from AFQ.tests.test_api import create_dummy_bids_path
from AFQ.api.group import GroupAFQ
import AFQ.data.fetch as afd
import pytest


def test_AFQ_custom_bundle_dict():
    bids_path = create_dummy_bids_path(3, 1)
    bundle_dict = abd.default18_bd()
    GroupAFQ(
        bids_path,
        preproc_pipeline="synthetic",
        bundle_info=bundle_dict)


def test_BundleDict():
    """
    Tests bundle dict
    """

    # test defaults
    afq_bundles = abd.default18_bd()

    assert len(afq_bundles) == 22

    # Arcuate Fasciculus
    afq_bundles = abd.default18_bd()["Left Arcuate", "Right Arcuate"]

    assert len(afq_bundles) == 2

    # Forceps Minor and Major
    afq_bundles = abd.default18_bd()["Forceps Major", "Forceps Minor"]

    assert len(afq_bundles) == 2

    # Test "custom" bundle
    afq_templates = afd.read_templates()
    afq_bundles = abd.BundleDict({
        "custom_bundle": {
            "include": [
                afq_templates["FA_L"],
                afq_templates["FP_R"]],
            "cross_midline": False}})
    afq_bundles.get("custom_bundle")

    assert len(afq_bundles) == 1

    # mispelled bundle that does not exist in afq templates
    with pytest.raises(
            ValueError,
            match=" is not in this BundleDict"):
        afq_bundles = abd.default18_bd()[
            "Left Vertical Occipital Quinticulus",
            "Right Vertical Occipital Quinticulus"]

    afq_bundles = abd.reco_bd(80)["VOF_L", "VOF_R"]
    assert len(afq_bundles) == 2



import logging
import dipy.tracking.utils as dtu
import numpy as np


logger = logging.getLogger('AFQ')


def get_percentile_threshold(mask, threshold):
    zero_mask = mask == 0
    mask[zero_mask] = np.nan
    new_threshold = np.nanpercentile(
        mask,
        100 - threshold)
    mask[zero_mask] = 0
    return new_threshold


def gen_seeds(seed_mask, seed_threshold,
              n_seeds, thresholds_as_percentages,
              random_seeds, rng_seed, affine):
    """
    Generate seeds for tracking. For parameters, see
    :func:`AFQ.tractography.tractography.track`.
    """
    logger.info("Generating Seeds...")
    if isinstance(n_seeds, int):
        if len(np.unique(seed_mask)) > 2:
            if thresholds_as_percentages:
                seed_threshold = get_percentile_threshold(
                    seed_mask, seed_threshold)
            seed_mask = seed_mask > seed_threshold
        if random_seeds:
            seeds = dtu.random_seeds_from_mask(seed_mask, seeds_count=n_seeds,
                                               seed_count_per_voxel=False,
                                               affine=affine,
                                               random_seed=rng_seed)
        else:
            seeds = dtu.seeds_from_mask(seed_mask,
                                        density=n_seeds,
                                        affine=affine)
    else:
        # If user provided an array, we'll use n_seeds as the seeds:
        seeds = n_seeds
    return seeds



import os.path as op
import os
import json


def write_json(fname, data):
    """
    Write data to JSON file.

    Parameters
    ----------
    fname : str
        Full path to the file to write.

    data : dict
        A dict containing the data to write.

    Returns
    -------
    None
    """
    with open(fname, 'w') as ff:
        json.dump(data, ff, default=lambda obj: "Not Serializable")


def read_json(fname):
    """
    Read data from a JSON file.

    Parameters
    ----------
    fname : str
        Full path to the data-containing file

    Returns
    -------
    dict
    """
    with open(fname, 'r') as ff:
        out = json.load(ff)
    return out


def drop_extension(path):
    base_fname = op.basename(path).split('.')[0]
    return path.split(base_fname)[0] + base_fname


def apply_cmd_to_afq_derivs(
        derivs_dir, base_fname, cmd="rm", exception_file_names=[], suffix="",
        dependent_on=None):
    if dependent_on is None:
        dependent_on_list = ["dwi", "trk", "rec", "prof"]
    elif dependent_on.lower() == "track":
        dependent_on_list = ["trk", "rec", "prof"]
    elif dependent_on.lower() == "recog":
        dependent_on_list = ["rec", "prof"]
    elif dependent_on.lower() == "prof":
        dependent_on_list = ["prof"]
    else:
        raise ValueError((
            "dependent_on must be one of "
            "None, 'track', 'recog', 'prof'."))

    for filename in os.listdir(derivs_dir):
        full_path = os.path.join(derivs_dir, filename)
        if os.path.isfile(full_path) or os.path.islink(full_path):
            if (full_path in exception_file_names)\
                    or (not full_path.startswith(base_fname))\
                    or filename.endswith("json"):
                continue
            sidecar_file = f'{drop_extension(full_path)}.json'
            if op.exists(sidecar_file):
                sidecar_info = read_json(sidecar_file)
                if "dependent" in sidecar_info\
                    and sidecar_info["dependent"]\
                        in dependent_on_list:
                    os.system(f"{cmd} {full_path} {suffix}")
                    os.system(f"{cmd} {sidecar_file} {suffix}")
            else:
                os.system(f"{cmd} {full_path} {suffix}")
        elif os.path.isdir(full_path):
            # other than ROIs, folders are dependent on everything
            if dependent_on is None or filename != "ROIs":
                os.system(f"{cmd} -r {full_path} {suffix}")



"""
====================
How pyAFQ uses BIDS
====================

The pyAFQ API relies heavily on the
`Brain Imaging Data Standard (BIDS) <https://bids-specification.readthedocs.io/en/stable/>`_,
a widely used standard for organizing and describing neuroimaging data. This
means that the software assumes that its inputs are organized according to the
BIDS specification and its outputs conform where possible with BIDS.

.. note::

    Derivatives of processing diffusion MRI are not currently fully
    described in the existing BIDS specification, but describing these
    is part of an ongoing effort. Wherever possible, we conform with
    the draft implementation of the BIDS DWI derivatives available
    `here <https://bids-specification.readthedocs.io/en/wip-derivatives/05-derivatives/05-diffusion-derivatives.html>`_

In this example, we will explore the use of BIDS in pyAFQ and see
how BIDS allows us to extend and provide flexibility to the users
of the software.
"""

import os
import os.path as op

import AFQ.api.bundle_dict as abd
from AFQ.api.group import GroupAFQ, get_afq_bids_entities_fname
import AFQ.data.fetch as afd


##########################################################################
# To interact with and query BIDS datasets, we use
#  `pyBIDS <https://bids-standard.github.io/pybids/>`_, which we import here:

import bids
from bids.layout import BIDSLayout

##########################################################################
# We start with some example data. The data we will use here is
# generated from the
# `Stanford HARDI dataset <https://purl.stanford.edu/ng782rw8378>`_.
# The call below fetches
# this dataset and organized it within the `~/AFQ_data` folder in the BIDS
# format.

afd.organize_stanford_data(clear_previous_afq="all")

##########################################################################
# After doing that, we should have a folder that looks like this:
#
# | stanford_hardi
# | ├── dataset_description.json
# | └── derivatives
# |     ├── freesurfer
# |     │   ├── dataset_description.json
# |     │   └── sub-01
# |     │       └── ses-01
# |     │           └── anat
# |     │               ├── sub-01_ses-01_T1w.nii.gz
# |     │               └── sub-01_ses-01_seg.nii.gz
# |     └── vistasoft
# |         ├── dataset_description.json
# |         └── sub-01
# |             └── ses-01
# |                 └── dwi
# |                     ├── sub-01_ses-01_dwi.bvals
# |                     ├── sub-01_ses-01_dwi.bvecs
# |                     └── sub-01_ses-01_dwi.nii.gz
#
# The top level directory (`stanford_hardi`) is our overall BIDS dataset folder.
# In many cases, this folder will include folders with raw data for each subject
# in the dataset. In this case, we do not include the raw data folders and only
# have the outputs of pipelines that were used to preprocess the data (e.g.,
# correct the data for subject motion, eddy currents, and so forth).
# In general, only the preprocessed diffusion data is required for pyAFQ to run.
# See the :doc:`"Organizing your data" </howto/usage/organizing>` section of the
# documentation for more details.
# In this case, one folder contains derivative of the Freesurfer software and
# another folder contains the DWI data that has been preprocessed with the
# Vistasoft software.
# pyAFQ provides facilities to segment tractography results obtained
# using other software as well. For example, we often use
# `qsiprep <https://qsiprep.readthedocs.io/en/latest/>`_ to preprocess
# our data and reconstruct tractographies with software such as
# `MRTRIX <https://www.mrtrix.org/>`_. Here, we will demonstrate how to use
# these reconstructions in the pyAFQ segmentation and tractometry pipeline
# We fetch this data and add it as a separate pipeline
# The following code will download a previously-created tractography and
# organize it by adding it to the BIDS dataset folder and renaming them to be
# BIDS-compliant (e.g., `sub-01_ses_01_dwi_tractography.trk`).

afd.fetch_stanford_hardi_tractography()

bids_path = op.join(op.expanduser('~'), 'AFQ_data', 'stanford_hardi')
tractography_path = op.join(bids_path, 'derivatives', 'my_tractography')
sub_path = op.join(tractography_path, 'sub-01', 'ses-01', 'dwi')

os.makedirs(sub_path, exist_ok=True)
os.rename(
    op.join(
        op.expanduser('~'),
        'AFQ_data',
        'stanford_hardi_tractography',
        'full_segmented_cleaned_tractography.trk'),
    op.join(
        sub_path,
        'sub-01_ses-01-dwi_tractography.trk'))

afd.to_bids_description(
    tractography_path,
    **{"Name": "my_tractography",
       "PipelineDescription": {"Name": "my_tractography"},
       "GeneratedBy": [{"Name": "my_tractography"}]})


###########################################################################
# After we do that, our dataset folder should look like this:
#
# | stanford_hardi
# | ├── dataset_description.json
# | └── derivatives
# |     ├── freesurfer
# |     │   ├── dataset_description.json
# |     │   └── sub-01
# |     │       └── ses-01
# |     │           └── anat
# |     │               ├── sub-01_ses-01_T1w.nii.gz
# |     │               └── sub-01_ses-01_seg.nii.gz
# |     ├── my_tractography
# |     |   ├── dataset_description.json
# |     │   └── sub-01
# |     │       └── ses-01
# |     │           └── dwi
# |     │               └── sub-01_ses-01-dwi_tractography.trk
# |     └── vistasoft
# |         ├── dataset_description.json
# |         └── sub-01
# |             └── ses-01
# |                 └── dwi
# |                     ├── sub-01_ses-01_dwi.bvals
# |                     ├── sub-01_ses-01_dwi.bvecs
# |                     └── sub-01_ses-01_dwi.nii.gz
#
# To explore the layout of these derivatives, we will initialize a
# :class:`BIDSLayout` class instance to help us see what is in this dataset

layout = bids.BIDSLayout(bids_path, derivatives=True)

##########################################################################
# Because there is no raw data in this BIDS layout (only derivatives),
# pybids will report that there are no subjects and sessions:

print(layout)

##########################################################################
# But a query on the derivatives will reveal the different derivatives that
# are stored here:

print(layout.derivatives)

##########################################################################
# We can use a :class:`bids.BIDSValidator` object to make sure that the
# files within our data set are BIDS-compliant. For example, we can
# extract the tractography derivatives part of our layout using:

my_tractography = layout.derivatives["my_tractography"]

##########################################################################
# This variable is also a BIDS layout object. This object has a ``get``
# method, which allows us to query and find specific items within the
# layout. For example, we can ask for files that have a suffix consistent
# with tractography results:

tractography_files = my_tractography.get(suffix='tractography')

##########################################################################
# Or ask for files that have a ``.trk`` extension:

tractography_files = my_tractography.get(extension='.trk')

##########################################################################
# In this case, both of these would produce the same result.

tractography_file = tractography_files[0]
print(tractography_file)

##########################################################################
# We can also get some more structured information about this file:

print(tractography_file.get_entities())


##########################################################################
# We can use a :class:`bids.BIDSValidator` class instance to validate that
# this file is compliant with the specification. Note that the validator
# requires that the filename be provided relative to the root of the BIDS
# dataset, so we have to split the string that contains the full path
# of the tractography to extract only the part that is relative to the
# root of the entire BIDS ``layout`` object:

tractography_full_path = tractography_file.path
tractography_relative_path = tractography_full_path.split(layout.root)[-1]

validator = bids.BIDSValidator()
print(validator.is_bids(tractography_relative_path))

##########################################################################
# Next, we specify the information we need to define the bundles that we are
# interested in segmenting. In this case, we are going to use a list of
# bundle names for the bundle info. These names refer to bundles for
# which we already have clear definitions of the information
# needed to segment them (e.g., waypoint ROIs and probability maps).
# For an example that includes custom definition of bundle info, see the
# `plot_callosal_tract_profile example <http://yeatmanlab.github.io/pyAFQ/auto_examples/plot_callosal_tract_profile.html>`_.

bundle_info = abd.default18_bd()[
    "Left Superior Longitudinal",
    "Right Superior Longitudinal",
    "Left Arcuate",
    "Right Arcuate",
    "Left Corticospinal",
    "Right Corticospinal",
    "Forceps Minor"]

##########################################################################
# Now, we can define our GroupAFQ object, pointing to the derivatives of the
# `'my_tractography'` pipeline as inputs. This is done by setting the
# `import_tract` key-word argument. We pass the
# `bundle_info` defined above. We also point to the preprocessed
# data that is in a `'dmriprep'` pipeline. Note that the pipeline name
# is not necessarily the name of the folder it is in; the pipeline name is
# defined in each pipeline's `dataset_description.json`. These data were
# preprocessed with 'vistasoft', so this is the pipeline we'll point to
# If we were using `'qsiprep'`, this is where we would pass that
# string instead. If we did that, AFQ would look for a derivatives
# folder called `'stanford_hardi/derivatives/qsiprep'` and find the
# preprocessed DWI data within it. Finally, to speed things up
# a bit, we also sub-sample the provided tractography. This is
# done by defining the segmentation_params dictionary input.
# To sub-sample to 10,000 streamlines, we define
# `'nb_streamlines' = 10000`.

my_afq = GroupAFQ(
    bids_path,
    preproc_pipeline='vistasoft',
    bundle_info=bundle_info,
    import_tract={
        "suffix": "tractography",
        "scope": "my_tractography"
    },
    segmentation_params={'nb_streamlines': 10000})

##########################################################################
# Finally, to run the segmentation and extract tract profiles, we call
# The `export_all` method. This creates all of the derivative outputs of
# AFQ within the 'stanford_hardi/derivatives/afq' folder.

my_afq.export_all()


##########################################################################
# A few common issues that can hinder BIDS from working properly are:
#
# 1. Faulty `dataset_description.json` file. You need to make sure that the
#    file contains the right names for the pipeline. See above for an example
#    of that.
# 2. File naming convention doesn't uniquely identify file with bids filters.


##########################################################################
# The outputs of AFQ are also BIDS compatible. However, we use some
# custom entities, which need to be loaded. Here we demonstrate how
# to load the afq entities with get_afq_bids_entities_fname()
# and then show all files with the key-value pair recogmethod-AFQ

layout = BIDSLayout(bids_path)
layout.add_derivatives(
    f'{bids_path}/derivatives/afq',
    config=['bids', 'derivatives', get_afq_bids_entities_fname()])
print(layout.get(recogmethod="AFQ", return_type="filename"))



"""
============================
Visualizing AFQ derivatives
============================

Visualizing the results of a pyAFQ analysis is useful because it allows us to
inspect the results of the analysis and to communicate the results to others.
The pyAFQ pipeline produces a number of different kinds of outputs, including
visualizations that can be used for quality control and for quick examination
of the results of the analysis.

However, when communicating the results of pyAFQ analysis, it is often useful
to have more specific control over the visualization that is produced. In
addition, it is often useful to have visualizations that are visually appealing
and striking. In this tutorial, we will use the `fury <https://fury.gl/>`_
library [1]_ to visualize outputs of pyAFQ as publication-ready figures.

"""

import os
import os.path as op
import nibabel as nib
import numpy as np

from dipy.io.streamline import load_trk
from dipy.tracking.streamline import transform_streamlines

from fury import actor, window
from fury.colormap import create_colormap

import AFQ.data.fetch as afd
from AFQ.viz.utils import PanelFigure

##############################################################################
# Get some data from HBN POD2
# ----------------------------
# The Healthy Brain Network Preprocessed Open Diffusion Derivatives (HBN POD2)
# is a collection of resources based on the Healthy Brain Network dataset
# [2, 3]_. HBN POD2 includes data derivatives - including pyAFQ derivatives -
# from more than 2,000 subjects. The data and the derivatives can be browsed at
# https://fcp-indi.s3.amazonaws.com/index.html#data/Projects/HBN/BIDS_curated/
#
# Here, we will visualize the results from one subject, together with their
# anatomy and using several variations. We start by downloading their
# pyAFQ-processed data using fetcher functions that download both the
# preprocessed data, as well as the pyAFQ-processed data (Note that this
# will take up about 1.75 GB of disk space):

afd.fetch_hbn_preproc(["NDARAA948VFH"])
study_path = afd.fetch_hbn_afq(["NDARAA948VFH"])[1]

deriv_path = op.join(
    study_path, "derivatives")

afq_path = op.join(
    deriv_path,
    'afq',
    'sub-NDARAA948VFH',
    'ses-HBNsiteRU')

bundle_path = op.join(afq_path,
                      'bundles')


#############################################################################
# Read data into memory
# ----------------------
# The bundle coordinates from pyAFQ are always saved in the reference frame of
# the diffusion data from which they are generated, so we need an image file
# with the dMRI coordinates as a reference for loading the data (we could also
# use `"same"` here).

fa_img = nib.load(op.join(afq_path,
                          'sub-NDARAA948VFH_ses-HBNsiteRU_acq-64dir_space-T1w_desc-preproc_dwi_model-DKI_FA.nii.gz'))
fa = fa_img.get_fdata()
sft_arc = load_trk(op.join(bundle_path,
                           'sub-NDARAA948VFH_ses-HBNsiteRU_acq-64dir_space-T1w_desc-preproc_dwi_space-RASMM_model-CSD_desc-prob-afq-ARC_L_tractography.trk'), fa_img)
sft_cst = load_trk(op.join(bundle_path,
                           'sub-NDARAA948VFH_ses-HBNsiteRU_acq-64dir_space-T1w_desc-preproc_dwi_space-RASMM_model-CSD_desc-prob-afq-CST_L_tractography.trk'), fa_img)

#############################################################################
# Transform into the T1w reference frame
# --------------------------------------
# Our first goal is to visualize the bundles with a background of the
# T1-weighted image, which provides anatomical context. We read in this data and
# transform the bundle coordinates, first into the RASMM common coordinate frame
# and then subsequently into the coordinate frame of the T1-weighted data (if
# you find this confusing, you can brush up on this topic in the
# `nibabel documentation <https://nipy.org/nibabel/coordinate_systems.html>`_).


t1w_img = nib.load(op.join(deriv_path,
                           'qsiprep/sub-NDARAA948VFH/anat/sub-NDARAA948VFH_desc-preproc_T1w.nii.gz'))
t1w = t1w_img.get_fdata()
sft_arc.to_rasmm()
sft_cst.to_rasmm()
arc_t1w = transform_streamlines(sft_arc.streamlines,
                                np.linalg.inv(t1w_img.affine))
cst_t1w = transform_streamlines(sft_cst.streamlines,
                                np.linalg.inv(t1w_img.affine))


#############################################################################
#
# .. note::
#   A virtual frame buffer is needed if you are running this example on
#   a machine that is not connected to a display ("headless"). If this is
#   the case, you can either set an environment variable called `XVFB` to `1`
#   or you can deindent the following code (and comment out the `if` statement)
#   to initialize the virtual frame buffer.

if os.environ.get("XVFB", False):
    print("Initializing XVFB")
    import xvfbwrapper
    from xvfbwrapper import Xvfb

    vdisplay = Xvfb()
    vdisplay.start()


#############################################################################
# Visualizing bundles with principal direction coloring
# -----------------------------------------------------
# The first visualization we will create will have the streamlines colored
# according to their direction. The color of each streamline will be RGB encoded
# according to the RAS of the average orientation of its segments. This is the
# default behavior in fury.
#
# Fury uses "actors" to render different kind of graphics. For the bundles, we
# will use the `line` actor. These objects are wrappers around the `vtkActor
# class <https://vtk.org/doc/nightly/html/classvtkActor.html>`_, so methods of
# that class (like `GetProperty()`) are available to use. We like to set the
# aesthetics of the streamlines, so that they are rendered as tubes and with
# slightly thicker line-width than the default. We create a function that sets
# these properties of the line actor via the `GetProperty` method. We will
# reuse this function later on, also setting the key-word arguments to the call
# to `actor.line`, but for now we use the default setting, which colors each
# streamline based on the RAS orientation, and we set the line width to 8.

def lines_as_tubes(sl, line_width, **kwargs):
    line_actor = actor.line(sl, **kwargs)
    line_actor.GetProperty().SetRenderLinesAsTubes(1)
    line_actor.GetProperty().SetLineWidth(line_width)
    return line_actor


arc_actor = lines_as_tubes(arc_t1w, 8)
cst_actor = lines_as_tubes(cst_t1w, 8)


#############################################################################
# Slicer actors
# -------------
# The anatomical image is rendered using `slicer` actors. These are actors that
# visualize one slice of a three dimensional volume. Again, we create a helper
# function that will slice a volume along the x, y, and z dimensions. This
# function returns a list of the slicers we want to include in our
# visualization. This can be one, two, or three slicers, depending on how many
# of {x,y,z} are set. If you are curious to understand what is going on in this
# function, take a look at the documentation for the
# :met:`actor.slicer.display_extent` method (hint: for every dimension you
# select on, you want the full extent of the image on the two *other* two
# dimensions). We call the function on the T1-weighted data, selecting the # x
# slice that is half-way through the x dimension of the image (`shape[0]`) and
# the z slice that is a third of a way through that x dimension of the image
# (`shape[-1]`).


def slice_volume(data, x=None, y=None, z=None):
    slicer_actors = []
    slicer_actor_z = actor.slicer(data)
    if z is not None:
        slicer_actor_z.display_extent(
            0, data.shape[0] - 1,
            0, data.shape[1] - 1,
            z, z)
        slicer_actors.append(slicer_actor_z)
    if y is not None:
        slicer_actor_y = slicer_actor_z.copy()
        slicer_actor_y.display_extent(
            0, data.shape[0] - 1,
            y, y,
            0, data.shape[2] - 1)
        slicer_actors.append(slicer_actor_y)
    if x is not None:
        slicer_actor_x = slicer_actor_z.copy()
        slicer_actor_x.display_extent(
            x, x,
            0, data.shape[1] - 1,
            0, data.shape[2] - 1)
        slicer_actors.append(slicer_actor_x)

    return slicer_actors


slicers = slice_volume(t1w, x=t1w.shape[0] // 2, z=t1w.shape[-1] // 3)

#############################################################################
# Making a `scene`
# -----------------
# The next kind of fury object we will be working with is a `window.Scene`
# object. This is the (3D!) canvas on which we are drawing the actors. We
# initialize this object and call the `scene.add` method to add the actors.

scene = window.Scene()

scene.add(arc_actor)
scene.add(cst_actor)
for slicer in slicers:
    scene.add(slicer)

#############################################################################
# Showing the visualization
# -------------------------
# If you are working in an interactive session, you can call::
#
#     window.show(scene, size=(1200, 1200), reset_camera=False)
#
# to see what the visualization looks like. This would pop up a window that will
# show you the visualization as it is now. You can interact with this
# visualization using your mouse to rotate the image in 3D, and mouse+ctrl or
# mouse+shift to pan and rotate it in plane, respectively. Use the scroll up and
# scroll down in your mouse to zoom in and out. Once you have found a view of
# the data that you like, you can close the window (as long as its open, it is
# blocking execution of any further commands in the Python interpreter!) and
# then you can query your scene for the "camera settings" by calling::
#
#     scene.camera_info()
#
# This will print out to the screen something like this::
#
#     # Active Camera
#        Position (238.04, 174.48, 143.04)
#        Focal Point (96.32, 110.34, 84.48)
#        View Up (-0.33, -0.12, 0.94)
#
# We can use the information we have gleaned to set the camera on subsequent
# visualization that use this scene object.


scene.set_camera(position=(238.04, 174.48, 143.04),
                 focal_point=(96.32, 110.34, 84.48),
                 view_up=(-0.33, -0.12, 0.94))

#############################################################################
# Record the visualization
# -------------------------
# If you are not working in an interactive session, or you have already figured
# out the camera settings that work for you, you can go ahead and record the
# image into a png file. We use a pretty high resolution here (2400 by 2400) so
# that we get a nice crisp image. That also means the file is pretty large.

out_folder = op.join(afd.afq_home, "VizExample")
os.makedirs(out_folder, exist_ok=True)
window.record(
    scene,
    out_path=op.join(out_folder, 'arc_cst1.png'),
    size=(2400, 2400))


############################################################################
# Setting bundle colors
# ---------------------
# The default behavior produces a nice image! But we often want to be able to
# differentiate streamlines based on the bundle to which they belong. For this,
# we will set the color of each streamline, based on the bundle. We often use
# the Tableau20 color palette to set the colors for the different bundles. The
# `actor.line` initializer takes `colors` as a keyword argument and one of the
# options to pass here is an RGB triplet, which will determine the color of all
# of the streamlines in that actor. We can get the Tableau20 RGB triplets from
# Matplotlib's colormap module. We initialize our bundle actors again and clear
# the scene before adding these new actors and adding back the slice actors. We
# then call `record` to create this new figure.

from matplotlib.cm import tab20
color_arc = tab20.colors[18]
color_cst = tab20.colors[2]

arc_actor = lines_as_tubes(arc_t1w, 8, colors=color_arc)
cst_actor = lines_as_tubes(cst_t1w, 8, colors=color_cst)

scene.clear()

scene.add(arc_actor)
scene.add(cst_actor)
for slicer in slicers:
    scene.add(slicer)

window.record(
    scene,
    out_path=op.join(out_folder, 'arc_cst2.png'),
    size=(2400, 2400))


#############################################################################
# Adding core bundles with tract profiles
# ---------------------------------------
# Next, we'd like to add a representation of the core of each bundle and plot
# the profile of some property along the length of the bundle. Another option
# for setting streamline colors is for each point along a streamline to have a a
# different color. Here, we will do this with only one streamline, which
# represents the core of the fiber -- it's median trajectory, and with a profile
# of the FA. This requires a few distinct steps. The first is adding another
# actor for each of the core fibers. We determine the core bundle as the median
# of the coordinates of the streamlines after each streamline is resampled to
# 100 equi-distant points. In the next step, we extract the FA profiles for each
# of the bundles, using the `afq_profile` function. Before we do this, we make
# sure that the streamlines are back in the voxel coordinate frame of the FA
# volume. The core bundle and FA profile are put together into a new line actor,
# which we initialize as a thick tube (line width of 40) and with a call to
# create a colormap from the FA profiles (we can choose which color-map to use
# here. In this case, we chose `'viridis'`). Note that because we are passing
# only one streamline into the line actor, we have to pass it inside of square
# brackets (`[]`). This is because the `actor.line` initializer expects a
# sequence of streamlines as input. Before plotting things again, we initialize
# our bundle line actors again. This time, we set the opacity of the lines to
# 0.1, so that they do not occlude the view of the core bundle visualizations.
#
# .. note::
#   In this case, the profile is FA, but you can use any sequence of 100
#   numbers in place of the FA profiles: group differences, p-values, etc.
#

from dipy.tracking.streamline import set_number_of_points
core_arc = np.median(np.asarray(set_number_of_points(arc_t1w, 100)), axis=0)
core_cst = np.median(np.asarray(set_number_of_points(cst_t1w, 100)), axis=0)


from dipy.stats.analysis import afq_profile
sft_arc.to_vox()
sft_cst.to_vox()
arc_profile = afq_profile(fa, sft_arc.streamlines, affine=np.eye(4))
cst_profile = afq_profile(fa, sft_cst.streamlines, affine=np.eye(4))

core_arc_actor = lines_as_tubes(
    [core_arc],
    40,
    colors=create_colormap(arc_profile, 'viridis')
)

core_cst_actor = lines_as_tubes(
    [core_cst],
    40,
    colors=create_colormap(cst_profile, 'viridis')
)

scene.clear()

arc_actor = lines_as_tubes(arc_t1w, 8, colors=color_arc, opacity=0.1)
cst_actor = lines_as_tubes(cst_t1w, 8, colors=color_cst, opacity=0.1)

scene.add(arc_actor)
scene.add(cst_actor)
for slicer in slicers:
    scene.add(slicer)
scene.add(core_arc_actor)
scene.add(core_cst_actor)

window.record(
    scene,
    out_path=op.join(out_folder, 'arc_cst3.png'),
    size=(2400, 2400))


#############################################################################
# Adding ROIs
# -----------
# Another element that we can add into the visualization are volume renderings
# of regions of interest. For example, the waypoint ROIs that were used to
# define the bundles. Here, we will add the waypoints that were used to define
# the arcuate fasciculus in this subject. We start by reading this data in.
# Because it is represented in the space of the diffusion data, it too needs to
# be resampled into the T1-weighted space, before being visualized. After
# resampling, we might have some values between 0 and 1 because of the
# interpolation from the low resolution of the diffusion into the high
# resolution of the T1-weighted. We will include in the volume rendering any
# values larger than 0. The main change from the previous visualizations is the
# adition of a `contour_from_roi` actor for each of the ROIs. We select another
# color from the Tableau 20 palette to represent this, and use an opacity of
# 0.5.
#
# .. note::
#   In this case, the surface rendered is of the waypoint ROIs, but very
#   similar code can be used to render other surfaces, provided a volume that
#   contains that surface. For example, the gray-white matter boundary could
#   be visualized provided an array with a binary representation of the volume #   enclosed in this boundary


from dipy.align import resample

waypoint1 = nib.load(
    op.join(
        afq_path,
        "ROIs", "sub-NDARAA948VFH_ses-HBNsiteRU_acq-64dir_space-T1w_desc-preproc_dwi_desc-ROI-ARC_L-1-include.nii.gz"))

waypoint2 = nib.load(
    op.join(
        afq_path,
        "ROIs", "sub-NDARAA948VFH_ses-HBNsiteRU_acq-64dir_space-T1w_desc-preproc_dwi_desc-ROI-ARC_L-2-include.nii.gz"))

waypoint1_xform = resample(waypoint1, t1w_img)
waypoint2_xform = resample(waypoint2, t1w_img)
waypoint1_data = waypoint1_xform.get_fdata() > 0
waypoint2_data = waypoint2_xform.get_fdata() > 0

scene.clear()

arc_actor = lines_as_tubes(arc_t1w, 8, colors=color_arc)
cst_actor = lines_as_tubes(cst_t1w, 8, colors=color_cst)

scene.add(arc_actor)
scene.add(cst_actor)
for slicer in slicers:
    scene.add(slicer)

surface_color = tab20.colors[0]

waypoint1_actor = actor.contour_from_roi(waypoint1_data,
                                         color=surface_color,
                                         opacity=0.5)

waypoint2_actor = actor.contour_from_roi(waypoint2_data,
                                         color=surface_color,
                                         opacity=0.5)


scene.add(waypoint1_actor)
scene.add(waypoint2_actor)


window.record(
    scene,
    out_path=op.join(out_folder, 'arc_cst4.png'),
    size=(2400, 2400))

#############################################################################
# Making a Figure out of many fury panels
# ---------------------------------------
# We can also make a figure that contains multiple panels, each of which
# contains a different visualization. This is useful for communicating the
# results of an analysis. Here, we will make a figure with four panels, using
# some of the visualizations we have already created. We will use some
# convenient methods from pyAFQ.

pf = PanelFigure(3, 2, 6, 9)
pf.add_img(op.join(out_folder, 'arc_cst1.png'), 0, 0)
pf.add_img(op.join(out_folder, 'arc_cst2.png'), 1, 0)
pf.add_img(op.join(out_folder, 'arc_cst3.png'), 0, 1)
pf.add_img(op.join(out_folder, 'arc_cst4.png'), 1, 1)
pf.format_and_save_figure(f"arc_cst_fig.png")

#############################################################################
#
# .. note::
#   If a virtual buffer was started before, it's a good idea to stop it.

if os.environ.get("XVFB", False):
    print("Stopping XVFB")
    vdisplay.stop()

#############################################################################
# References
# ----------
# .. [1] Garyfallidis et al., (2021). FURY: advanced scientific visualization.
#    Journal of Open Source Software, 6(64), 3384,
#    https://doi.org/10.21105/joss.03384
#
# .. [2] Alexander LM, Escalera J, Ai L, et al. An open resource for
#     transdiagnostic research in pediatric mental health and learning
#     disorders. Sci Data. 2017;4:170181.
#
# .. [3] Richie-Halford A, Cieslak M, Ai L, et al. An analysis-ready and
#     quality controlled resource for pediatric brain white-matter research.
#     Scientific Data. 2022;9(1):1-27.
